import { BackupId, DatabaseId, InstanceId, IssueId, PrincipalId, ProjectId } from "./id";
import { Pipeline, PipelineCreate } from "./pipeline";
import { Principal } from "./principal";
import { Project } from "./project";
import { MigrationType } from "./instance";

type IssueTypeGeneral = "bb.issue.general";

type IssueTypeDatabase =
  | "bb.issue.database.create"
  | "bb.issue.database.grant"
  | "bb.issue.database.schema.update";

type IssueTypeDataSource = "bb.issue.data-source.request";

export type IssueType =
  | IssueTypeGeneral
  | IssueTypeDatabase
  | IssueTypeDataSource;

export type IssueStatus = "OPEN" | "DONE" | "CANCELED";

export type CreateDatabaseContext = {
  instanceId: InstanceId;
  databaseName: string;
  characterSet: string;
  collation: string;
  backupId: BackupId;
  backupName: string;
};

export type UpdateSchemaDetail = {
  databaseId: DatabaseId;
  statement: string;
  rollbackStatement: string;
};

export type UpdateSchemaContext = {
  migrationType: MigrationType;
  updateSchemaDetailList: UpdateSchemaDetail[];
};

export type EmptyContext = {
};

export type IssueCreateContext = 
  | CreateDatabaseContext
  | UpdateSchemaContext
  | EmptyContext;

export type IssuePayload = { [key: string]: any };

export type Issue = {
  id: IssueId;

  // Related fields
  project: Project;
  pipeline: Pipeline;

  // Standard fields
  creator: Principal;
  createdTs: number;
  updater: Principal;
  updatedTs: number;

  // Domain specific fields
  name: string;
  status: IssueStatus;
  type: IssueType;
  description: string;
  assignee: Principal;
  subscriberIdList: PrincipalId[];
  payload: IssuePayload;
};

export type IssueCreate = {
  // Related fields
  projectId: ProjectId;
  pipeline: PipelineCreate;

  // Domain specific fields
  name: string;
  type: IssueType;
  description: string;
  assigneeId: PrincipalId;
  rollbackIssueId?: IssueId;
  createContext: IssueCreateContext;
  payload: IssuePayload;
};

export type IssuePatch = {
  // Domain specific fields
  name?: string;
  description?: string;
  assigneeId?: PrincipalId;
  payload?: IssuePayload;
};

export type IssueStatusPatch = {
  // Domain specific fields
  status: IssueStatus;
  comment?: string;
};

export type IssueStatusTransitionType = "RESOLVE" | "CANCEL" | "REOPEN";

export interface IssueStatusTransition {
  type: IssueStatusTransitionType;
  to: IssueStatus;
  buttonName: string;
  buttonClass: string;
}

export const ISSUE_STATUS_TRANSITION_LIST: Map<
  IssueStatusTransitionType,
  IssueStatusTransition
> = new Map([
  [
    "RESOLVE",
    {
      type: "RESOLVE",
      to: "DONE",
      buttonName: "issue.status-transition.dropdown.resolve",
      buttonClass: "btn-success",
    },
  ],
  [
    "CANCEL",
    {
      type: "CANCEL",
      to: "CANCELED",
      buttonName: "issue.status-transition.dropdown.cancel",
      buttonClass: "btn-normal",
    },
  ],
  [
    "REOPEN",
    {
      type: "REOPEN",
      to: "OPEN",
      buttonName: "issue.status-transition.dropdown.reopen",
      buttonClass: "btn-normal",
    },
  ],
]);

// The first transition in the list is the primary action and the rests are
// the normal action. For now there are at most 1 primary 1 normal action.
export const CREATOR_APPLICABLE_ACTION_LIST: Map<
  IssueStatus,
  IssueStatusTransitionType[]
> = new Map([
  ["OPEN", ["CANCEL"]],
  ["DONE", ["REOPEN"]],
  ["CANCELED", ["REOPEN"]],
]);

export const ASSIGNEE_APPLICABLE_ACTION_LIST: Map<
  IssueStatus,
  IssueStatusTransitionType[]
> = new Map([
  ["OPEN", ["RESOLVE", "CANCEL"]],
  ["DONE", ["REOPEN"]],
  ["CANCELED", ["REOPEN"]],
]);
